import warnings
import numpy as np
from ADPYNE.AutoDiff import AutoDiff
from ADPYNE.Dual import Dual

#-------------------BASE TRIG FUNCTIONS-------------------#
# Sine function


def sin(x):
	''' Compute the sine of an AutoDiff object and its derivative.
		Compute the sine of Dual Number

	INPUTS
	======
	x: an AutoDiff object

	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.

	EXAMPLES
	========
	>>> x = AutoDiff(0.5, 2.0, 1.0)
	>>> myAutoDiff = sin(x)
	>>> myAutoDiff.val
	0.479425538604
	>>> myAutoDiff.der
	1.75516512378
	>>> myAutoDiff.jacobian
	0.87758256189

	'''
	try:
		new_val = np.sin(x.val)
		new_der = np.cos(x.val) * x.der
		new_jacobian = np.cos(x.val) * x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.sin(x.Real), x.Dual * np.cos(x.Real))
		except AttributeError:
			try:
				return Dual(sin(x.Real), x.Dual * cos(x.Real))
			except AttributeError:
				return np.sin(x)

# Cosine function


def cos(x):
	''' Compute the cosine of an AutoDiff object and its derivative.

	INPUTS
	======
	x: an AutoDiff object

	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.

	EXAMPLES
	========
	>>> x = AutoDiff(0.5, 2.0, 1.0)
	>>> myAutoDiff = cos(x)
	>>> myAutoDiff.val
	0.87758256189
	>>> myAutoDiff.der
	-0.958851077208
	>>> myAutoDiff.jacobian
	-0.479425538604

	'''
	try:
		new_val = np.cos(x.val)
		new_der = -1.0 * np.sin(x.val) * x.der
		new_jacobian = -1.0 * np.sin(x.val) * x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.cos(x.Real), x.Dual * -np.sin(x.Real))
		except AttributeError:
			try:
				return Dual(cos(x.Real), x.Dual * -sin(x.Real))
			except AttributeError:
				return np.cos(x)

# Tangent function


def tan(x):
	''' Compute the tangent of an AutoDiff object and its derivative.

	INPUTS
	======
	x: an AutoDiff object

	'''
	try:
		# Value and derivative undefined when divisible by pi/2 but not pi
		# To make sure the asymptotes are undefined:
		if x.val%(np.pi/2)==0 and x.val%np.pi!=0:
			new_val = np.nan
			new_der = np.nan
			new_jacobian = np.nan
			warnings.warn('Undefined at value', RuntimeWarning)
		else:
			new_val = np.tan(x.val)
			new_der = x.der / (np.cos(x.val)**2.0)
			new_jacobian = x.jacobian / (np.cos(x.val)**2.0)
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			if x.Real%(np.pi/2)==0 and x.Real%np.pi!=0:
				ans = Dual(np.nan,np.nan)
				warnings.warn('Undefined at value', RuntimeWarning)
				return ans
			else:
				return Dual(np.tan(x.Real), x.Dual / (np.cos(x.Real))**2)
		except AttributeError:
			try:
				if x.Real%(np.pi/2)==0 and x.Real%np.pi!=0:
					ans = Dual(np.nan,np.nan)
					warnings.warn('Undefined at value', RuntimeWarning)
					return ans
				else:
					# return Dual(tan(x.Real), x.Dual / (cos(x.Real))**2)
					return sin(x)/cos(x)
			except AttributeError:
				if x%(np.pi/2)==0 and x%np.pi!=0:
					warnings.warn('Undefined at value', RuntimeWarning)
					return np.nan
				else:
					return np.tan(x)

#-------------------INVERSE TRIG FUNCTIONS-------------------#
# arc sin
def arcsin(X):
	''' Compute the arcsin of an AutoDiff object and its derivative.

		INPUTS
		======
		X: an AutoDiff object or constant

		RETURNS
		=======
		A new AutoDiff object or scalar with calculated value and derivative.

		EXAMPLES
		========
		>>> X = AutoDiff(0.5, 2)
		>>> arcsinAutoDiff = arcsin(X)
		>>> arcsinAutoDiff.val
		0.5235987755982988
		>>> arcsinAutoDiff.der
		2.3094010767585034
		>>> arcsinAutoDiff.jacobian
		1.1547005383792517
		'''

	try:
		# Is another ADT
		new_val = np.arcsin(X.val) 
		new_der = (1/np.sqrt(1-X.val**2))*X.der 
		new_jacobian = (1/np.sqrt(1-X.val**2))*X.jacobian 
		
		return AutoDiff(new_val, new_der, X.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.arcsin(X.Real), X.Dual/np.sqrt(1-X.Real**2))
		except AttributeError:
			try:
				# return Dual(arcsin(X.Real), X.Dual/sqrt(1-X.Real**2))
				return Dual(arcsin(X.Real), (X.Dual*(1-X.Real**2)**-0.5))
			except AttributeError:
			# Constant
				return_val = np.arcsin(X)
				return return_val

# arc cosine


def arccos(X):
	''' Compute the arccos of an AutoDiff object and its derivative.

		INPUTS
		======
		X: an AutoDiff object or constant

		RETURNS
		=======
		A new AutoDiff object or scalar with calculated value and derivative.

		EXAMPLES
		========
		>>> X = AutoDiff(0.5, 2)
		>>> arccosAutoDiff = arccos(X)
		>>> arccosAutoDiff.val
		1.0471975511965976
		>>> arccosAutoDiff.der
		-2.3094010767585034
		>>> arccosAutoDiff.jacobian
		-1.1547005383792517
		'''
	try:
		# Is another ADT
		new_val = np.arccos(X.val) #if (-1 <= X.val and X.val <= 1) else np.nan
		new_der = (-1/np.sqrt(1-X.val**2))*X.der #if (-1 < X.val and X.val < 1) else np.nan
		new_jacobian = (-1/np.sqrt(1-X.val**2))*X.jacobian #if (-1 < X.val and X.val < 1) else np.nan

		return AutoDiff(new_val, new_der, X.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.arccos(X.Real), -X.Dual/np.sqrt(1-X.Real**2))		
		except AttributeError:
			try:
				return Dual(arccos(X.Real), -X.Dual/sqrt(1-X.Real**2))
			except AttributeError:
			# Constant
				return_val = np.arccos(X)
				return return_val

# arc tangent
def arctan(X):
	''' Compute the arctan of an AutoDiff object and its derivative.

		INPUTS
		======
		X: an AutoDiff object or constant

		RETURNS
		=======
		A new AutoDiff object or scalar with calculated value and derivative.

		EXAMPLES
		========
		>>> X = AutoDiff(3, 2)
		>>> arctanAutoDiff = arctan(X)
		>>> arctanAutoDiff.val
		1.2490457723982544
		>>> arctanAutoDiff.der
		0.2
		>>> arctanAutoDiff.jacobian
		0.1	
		'''

	try:
		# Is another ADT
		new_val = np.arctan(X.val)
		new_der = (1/(1+X.val**2))*X.der
		new_jacobian = (1/(1+X.val**2))*X.jacobian
		return AutoDiff(new_val, new_der, X.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.arctan(X.Real), X.Dual/(1+X.Real**2))		
		except AttributeError:
			try:
				return Dual(arctan(X.Real), X.Dual/(1+X.Real**2))
			except AttributeError:
			# Constant
				return_val = np.arctan(X)
				return return_val

#-------------------HYPERBOLIC TRIG FUNCTIONS-------------------#
# hyperbolic sin
def sinh(X):
	''' Compute the sinh of an AutoDiff object and its derivative.
	INPUTS
	======
	X: an AutoDiff object
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	EXAMPLES
	========
	>>> X = AutoDiff(0.5, 2, 1)
	>>> sinhAutoDiff = sinh(X)
	>>> sinhAutoDiff.val
	0.5210953054937474
	>>> sinhAutoDiff.der
	2.2552519304127614
	>>> sinhAutoDiff.jacobian
	1.1276259652063807
	'''
	try:
		val = np.sinh(X.val)
		der = np.cosh(X.val)*X.der
		jacobian = np.cosh(X.val)*X.jacobian
		return AutoDiff(val, der, X.n, 0, jacobian)
	except AttributeError:
		try:
			return Dual(np.sinh(X.Real), X.Dual*np.cosh(X.Real))		
		except AttributeError:
			try:
				return Dual(sinh(X.Real), X.Dual*cosh(X.Real))
			except AttributeError:
			# Constant
				return_val = np.sinh(X)
				return return_val

# hyperbolic cos


def cosh(X):
	''' Compute the cosh of an AutoDiff object and its derivative.
	INPUTS
	======
	X: an AutoDiff object
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	EXAMPLES
	========
	>>> X = AutoDiff(0.5, 2, 1)
	>>> coshAutoDiff = cosh(X)
	>>> coshAutoDiff.val
	1.1276259652063807
	>>> coshAutoDiff.der
	1.0421906109874948
	>>> coshAutoDiff.jacobian
	0.5210953054937474
	'''
	try:
		val = np.cosh(X.val)
		der = np.sinh(X.val)*X.der
		jacobian = np.sinh(X.val)*X.jacobian
		return AutoDiff(val, der, X.n, 0, jacobian)
	except AttributeError:
		try:
			return Dual(np.cosh(X.Real), X.Dual*np.sinh(X.Real))		
		except AttributeError:
			try:
				return Dual(cosh(X.Real), X.Dual*sinh(X.Real))
			except AttributeError:
			# Constant
				return_val = np.cosh(X)
				return return_val

# hyperbolic tan


def tanh(X):
	''' Compute the tanh of an AutoDiff object and its derivative.
	INPUTS
	======
	X: an AutoDiff object
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	EXAMPLES
	========
	>>> X = AutoDiff(0.5, 2, 1)
	>>> tanhAutoDiff = tanh(X)
	>>> tanhAutoDiff.val
	0.46211715726000974
	>>> tanhAutoDiff.der
	1.572895465931855
	>>>tanhAutoDiff.jacobian
	0.7864477329659275
	'''
	try:
		val = np.tanh(X.val)
		der = 1/(np.cosh(X.val)**2)*X.der
		jacobian = 1/(np.cosh(X.val)**2)*X.jacobian
		return AutoDiff(val, der, X.n, 0, jacobian)
	except AttributeError:
		try:
			return Dual(np.tanh(X.Real), X.Dual/(np.cosh(X.Real)**2))		
		except AttributeError:
			try:
				X.Real
				return sinh(X)/cosh(X)
			except AttributeError:
			# Constant
				return_val = np.tanh(X)
				return return_val

#-------------------ARC HYPERBOLIC TRIG FUNCTIONS-------------------#
# hyperbolic arcsin


def arcsinh(x):
	''' Compute the hyperbolic arc sine of an AutoDiff object and its derivative.
	
	INPUTS
	======
	x: an AutoDiff object
	
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	
	EXAMPLES
	========
	>>> x = AutoDiff(0.5, 2, 1)
	>>> myAutoDiff = arcsinh(x)
	>>> myAutoDiff.val
	2.3124383412727525
	>>> myAutoDiff.der
	0.39223227027
	>>> myAutoDiff.jacobian
	0.19611613513818404
	
	'''
	try:
		new_val = np.arcsinh(x.val)
		new_der = ((1)/np.sqrt(x.val**2 + 1))*x.der
		new_jacobian = ((1)/np.sqrt(x.val**2 + 1))*x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.arcsinh(x.Real), x.Dual/np.sqrt((x.Real**2)+1))		
		except AttributeError:
			try:
				return Dual(arcsinh(x.Real), (x.Dual*(1+x.Real**2)**-0.5))
			except AttributeError:
			# Constant
				return_val = np.arcsinh(x)
				return return_val

# hyperbolic arc cosine


def arccosh(x):
	''' Compute the hyperbolic arc cosine of an AutoDiff object and its derivative.
	
	INPUTS
	======
	x: an AutoDiff object
	
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	
	EXAMPLES
	========
	>>> x = AutoDiff(1.1, 2)
	>>> myAutoDiff = arccosh(x)
	>>> myAutoDiff.val
	0.4435682543851154
	>>> myAutoDiff.der
	(2/np.sqrt(1.1**2 - 1))
	>>> myAutoDiff.jacobian
	(1/np.sqrt(1.1**2 - 1))
	
	'''
	try:
		new_val = np.arccosh(x.val)
		# Derivative of arccosh is only defined when x > 1
		new_der = ((1)/np.sqrt(x.val**2 - 1))*x.der  # if x.val > 1 else None
		new_jacobian = ((1)/np.sqrt(x.val**2 - 1))*x.jacobian  # if x.val > 1 else None
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.arccosh(x.Real), x.Dual/np.sqrt((x.Real**2)-1))		
		except AttributeError:
			try:
				return Dual(arccosh(x.Real), (x.Dual*((x.Real**2)-1)**-0.5))
			except AttributeError:
			# Constant
				return_val = np.arccosh(x)
				return return_val

# hyperbolic arc tangent


def arctanh(x):
	''' Compute the hyperbolic arc tangent of an AutoDiff object and its derivative.
	
	INPUTS
	======
	x: an AutoDiff object
	
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	
	EXAMPLES
	========
	>>> x = AutoDiff(0.5, 2)
	>>> myAutoDiff = arctanh(x)
	>>> myAutoDiff.val
	0.5493061443340548
	>>> myAutoDiff.der
	2/(1-(0.5)**2)
	>>> myAutoDiff.jacobian
	1/(1-(0.5)**2)
	
	'''
	try:
		new_val = np.arctanh(x.val)
		new_der = ((1)/(1-x.val**2))*x.der
		new_jacobian = ((1)/(1-x.val**2))*x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			if(np.abs(x.Real)==1):
				real = np.inf
				dual = np.inf
				warnings.warn('Undefined at value', RuntimeWarning)
			else:
				real = np.arctanh(x.Real)
				dual = x.Dual/(1-x.Real**2)
			return Dual(real, dual)	
		except AttributeError:
			try:
				return Dual(arctanh(x.Real), x.Dual/(1-x.Real**2))
			except AttributeError:
			# Constant
				return_val = np.arctanh(x)
				return return_val


#--------------------------EXPONENT FAMILY----------------------------#
# Exponential


def exp(x):
	''' Compute the exponential of an AutoDiff object and its derivative.
	
	INPUTS
	======
	x: an AutoDiff object
	
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	
	EXAMPLES
	========
	>>> x = AutoDiff(10, 2)
	>>> myAutoDiff = exp(x)
	>>> myAutoDiff.val
	22026.465794806718
	>>> myAutoDiff.der
	2*22026.465794806718
	>>> myAutoDiff.jacobian
	22026.465794806718	
	'''
	try:
		new_val = np.exp(x.val)
		new_der = np.exp(x.val) * x.der
		new_jacobian = np.exp(x.val) * x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.exp(x.Real), x.Dual*np.exp(x.Real))		
		except AttributeError:
			try:
				return Dual(exp(x.Real), x.Dual*exp(x.Real))
			except AttributeError:
			# Constant
				return_val = np.exp(x)
				return return_val

# natural log


def log(x):
	''' Compute the natural log of an AutoDiff object and its derivative.
	
	INPUTS
	======
	x: an AutoDiff object
	
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	
	EXAMPLES
	========
	x = AutoDiff(4, 2)
	>>> myAutoDiff = log(x)
	>>> myAutoDiff.val
	1.3862943611198906
	>>> myAutoDiff.der
	0.5
	>>> myAutoDiff.jacobian
	0.25
	
	'''
	try:
		new_val = np.log(x.val)
		# Derivative not defined when x = 0
		new_der = (1/(x.val*np.sum(1)))*x.der # if x.val != 0 else None
		new_jacobian = (1/(x.val*np.sum(1)))*x.jacobian # if x.val != 0 else None
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			if(x.Real==0):
				real = -np.inf
				dual = np.inf
			else:
				real = np.log(x.Real)
				dual = x.Dual/x.Real
			return Dual(real, dual)		
		except AttributeError:
			try:
				return Dual(log(x.Real), x.Dual/x.Real)
			except AttributeError:
			# Constant
				return_val = np.log(x)
				return return_val

# log base 10
def log10(x):
	''' Compute the natural log of an AutoDiff object and its derivative.
	
	INPUTS
	======
	x: an AutoDiff object
	
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	
	EXAMPLES
	========
	>>> X = AutoDiff(0.5, 2, 1)
	>>> myAutoDiff = log(X)
	>>> myAutoDiff.val
	-0.3010299956639812
	>>> myAutoDiff.der
	1.737177927613007
	>>>myAutoDiff.jacobian
	0.8685889638065035
	
	'''
	try:
		new_val = np.log10(x.val)
		# Derivative not defined when x = 0
		new_der = (1/(x.val*np.log(10)))*x.der
		new_jacobian = (1/(x.val*np.log(10)))*x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			real = np.log10(x.Real)
			dual = x.Dual/(x.Real*np.log(10))
			return Dual(real, dual)		
		except AttributeError:
			try:
				return Dual(log(x.Real)/np.log(10), x.Dual/(x.Real*(np.log(10))))
			except AttributeError:
			# Constant
				return_val = np.log10(x)
				return return_val

# Square Root


def sqrt(x):
	''' Compute the square root an AutoDiff object and its derivative.

	INPUTS
	======
	x: an AutoDiff object

	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.

	EXAMPLES
	========
	>>> x = AutoDiff(np.array([[5]]).T, np.array([[1]]), 1, 1)
	>>> myAutoDiff = sqrt(x)
	>>> myAutoDiff.val
	2.2360679775
	>>> myAutoDiff.der
	0.2236068

	'''
	try:
		new_val = np.sqrt(x.val)
		new_der = 0.5 * x.val ** (-0.5) * x.der
		new_jacobian = 0.5 * x.val ** (-0.5) * x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
			try:
				if x.Real < 0.0:
					warnings.warn('Undefined at value', RuntimeWarning)
					dual=np.nan
				
				elif(x.Real==0):
					warnings.warn('Undefined at value', RuntimeWarning)
					dual = np.inf
				
				else:
					dual = 0.5 * x.Real ** (-0.5) * x.Dual
				
				real = np.sqrt(x.Real)
				return Dual(real, dual)
			except AttributeError:
				if x < 0.0:
					warnings.warn('Undefined at value', RuntimeWarning)
					return np.nan
				else:
					return np.sqrt(x)

# log base
	
def logbase(x,base):
	''' Compute any log base of an AutoDiff object and its derivative.
	
	INPUTS
	======
	x: an AutoDiff object
	
	RETURNS
	=======
	A new AutoDiff object with calculated value and derivative.
	
	EXAMPLES
	========
	x = AutoDiff(2, 2)
	>>> myAutoDiff = logbase(x,7)
	>>> myAutoDiff.val
	0.35620719
	>>> myAutoDiff.der
	0.51389834
	>>> myAutoDiff.jacobian
	0.25694917
	
	'''
	try:
		new_val = np.log(x.val)/np.log(base)
		# Derivative not defined when x = 0
		new_der = (1/(x.val*np.log(base)))*x.der
		new_jacobian = (1/(x.val*np.log(base)))*x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			return Dual(np.log(x.Real)/np.log(base), x.Dual/(x.Real*np.log(base)))		
		except AttributeError:
			try:
				return Dual(log(x.Real)/np.log(base), x.Dual/(x.Real*(np.log(base))))
			except AttributeError:
			# Constant
				return_val = np.log(x)/np.log(base)
				return return_val

def logistic(x):
	''' Compute logistic function for AutoDiff or Dual object.
	
	INPUTS
	======
	x: an AutoDiff object or Dual object
	
	RETURNS
	=======
	A new AutoDiff or Dual object with calculated value and derivative.
	
	
	'''
	try:
		f_l = (1/(1+np.exp(-x.val)))
		new_val = f_l
		new_der = (1 - f_l)*f_l*x.der
		new_jacobian = (1 - f_l)*f_l*x.jacobian
		return AutoDiff(new_val, new_der, x.n, 0, new_jacobian)
	except AttributeError:
		try:
			f_l = (1/(1 + np.exp(-x.Real)))
			return Dual(f_l, (1 - f_l)*f_l*x.Dual)		
		except AttributeError:
			try:
				return Dual(logistic(x.Real), (1 - logistic(x.Real))*logistic(x.Real)*x.Dual)
			except AttributeError:
			# Constant
				return_val = (1/(1+np.exp(-x)))
				return return_val



import numpy as np

class Hessian():
	def __init__(self, f):
		self.value = f.Real.Real.Real

		fd1 = f.Dual[0].Real.Real[0].Real.Real
		fd2 = f.Dual[1].Real.Real[1].Real.Real
		self.firstDer = np.array([fd1,fd2])
		
		hxx = f.Dual[0].Dual[0].Real[0].Real
		hyy = f.Dual[0].Real.Real[1].Dual[0].Real[1].Real
		hxy = f.Dual[0].Dual[0].Real[1].Real
		hyx = hxy
		self.hessian = np.array([[hxx,hxy],[hyx,hyy]])


	def __str__(self):
		return "{} val\n\n{} der\n\n{} hess".format(self.value,self.firstDer,self.hessian)


import numpy as np

class AutoDiff():
	'''
	An auto-differentiation object for scalar and vector functions
	'''

	def __init__(self, eval_value, der_value, n=1, k=1, jacobian_value = np.array([[None]])):
		'''
		INPUTS
		======
		eval_value: 	value to be evaluated at
		der_value: 		evaluated value of the derivative
		n:				number of input variables the final function will use
		k:				denotes that the autodiff uses the kth input variable in a vector of 1 to n variables
		jacob_value: 	denotes the evaluted value of the jacobian

		RETURNS
		=======
		An AutoDiff object with calculated value, derivative, and jacobian.

		EXAMPLES
		========
		>>> myAutoDiff = AutoDiff(3, 2)
		>>> myAutoDiff.val
		3
		>>> myAutoDiff.der
		2
		>>> myAutoDiff.jacobian
		1
		'''
		# Convert int or float to array
		self.val = self._convertNonArray(eval_value, k)
		self.jacobian = self._calcJacobian(k, n, jacobian_value)
		self.der = self._calcDerivative(der_value, k)
		self.n = n

	def _convertNonArray(self, value, k):
		# try:
		# 	value.shape
		# 	return value
		# except:
		# 	return np.array([[value]])
		if k != 0:
			try:
				return np.array(value).reshape(len(value), 1)
			except:
				return np.array([[value]])
		else:
			return value

	def _calcJacobian(self, k, n, jacobian_value):
		if np.all(np.equal(jacobian_value, None)):
			if k != 0:
				rows = self.val.shape[0]
				seed = np.zeros([rows, n])
				seed[:, k-1] = 1
				return seed
		else:
			return jacobian_value

	def _calcDerivative(self, der_value, k):
		if k != 0:
			jacobian = (np.array([self.jacobian[0]]))
			der = self._convertNonArray(der_value, k)
			return np.dot(der, jacobian) * self.jacobian**0
		else:
			return self._convertNonArray(der_value, k)

	def __str__(self):
		return "Value:\n{}\nDerivative:\n{}\nJacobian:\n{}\n".format(self.val, self.der, self.jacobian)

	def __repr__(self):
		return "Value:\n{}\nDerivative:\n{}\nJacobian:\n{}\n".format(self.val, self.der, self.jacobian)

	def __add__(self, other):
		try:
			# If  AutoDiff of same variable, values and derivatives should both just add
			return AutoDiff(self.val + other.val, self.der + other.der, self.n, 0, self.jacobian + other.jacobian)
		except AttributeError:
			# If trying to add a constant to AutoDiff, only add values. Constant has derivative of 0 so no addition needed.
			return AutoDiff(self.val + other, self.der, self.n, 0, self.jacobian)

	# Account for reverse addition
	def __radd__(self, other):
		try:
			return AutoDiff(self.val + other.val, self.der + other.der, self.n, 0, self.jacobian + other.jacobian)
		except AttributeError:
			return AutoDiff(self.val + other, self.der, self.n, 0, self.jacobian)

	def __sub__(self, other):
		try:
			# If  AutoDiff of same variable, values and derivatives should both just add
			return AutoDiff(self.val - other.val, self.der - other.der, self.n, 0, self.jacobian - other.jacobian)
		except AttributeError:
			# If trying to add a constant to AutoDiff, only add values. Constant has derivative of 0 so no subtraction needed.
			return AutoDiff(self.val - other, self.der, self.n, 0, self.jacobian)

	# Account for reverse subtraction
	def __rsub__(self, other):
		try:
			return AutoDiff(other.val - self.val, other.der - self.der, self.n, 0, other.jacobian - self.jacobian)
		except AttributeError:
			return AutoDiff(other - self.val, self.der, self.n, 0, self.jacobian)

	def __mul__(self, other):
		try:
			# Use product rule
			return AutoDiff(self.val * other.val, self.val * other.der + self.der * other.val, self.n, 0, self.val * other.jacobian + self.jacobian * other.val)
		except AttributeError:
			return AutoDiff(self.val * other, self.der * other, self.n, 0, self.jacobian * other)

	# Account for reverse multiplication
	def __rmul__(self, other):
		try:
			# Use product rule
			return AutoDiff(self.val * other.val, self.val * other.der + self.der * other.val, self.n, 0, self.val * other.jacobian + self.jacobian * other.val)
		except AttributeError:
			return AutoDiff(self.val * other, self.der * other, self.n, 0, self.jacobian * other)

	def __truediv__(self, other):
		try:
			# Use quotient rule
			return AutoDiff(self.val / other.val, (self.der * other.val - self.val * other.der)/(other.val**2), self.n, 0, (self.jacobian * other.val - self.val * other.jacobian)/(other.val**2))
		except AttributeError:
			return AutoDiff(self.val / other, self.der / other, self.n, 0, self.jacobian / other)

	# Account for reverse true division
	def __rtruediv__(self, other):
		try:
			# Use quotient rule
			# other/self
			return AutoDiff(other.val / self.val, (self.val * other.der - other.val * self.der)/(self.val**2), self.n, 0, (other.jacobian * self.val - other.val * self.jacobian)/(self.val**2) )
		except AttributeError:
			return AutoDiff(other / self.val, (self.val * 0 - other * self.der)/(self.val**2), self.n, 0, (self.val * 0 - other * self.jacobian)/(self.val**2))

	def __pow__(self, other):
		# Convert to float so that negative integers will work
		other = float(other) if type(other)==int else other
		try:
			return AutoDiff(self.val**other.val, other.val * (self.val ** (other.val-1)) * self.der + (self.val**other.val) *np.log(np.abs(self.val)) * other.der, self.n, 0, other.val * (self.val**(other.val-1)) * self.jacobian + (self.val**other.val * np.log(np.abs(self.val)) * other.jacobian))
		except AttributeError:
			return AutoDiff(self.val**other, other * (self.der) * self.val**(other-1), self.n, 0, other * (self.jacobian) * self.val**(other-1))

	def __rpow__(self, other):
		try:
			return AutoDiff(other.val**self.val, other.val * (self.val ** (other.val-1)) * self.der + (self.val**other.val) *np.log(np.abs(self.val)) * other.der, self.n, 0, other.val * (self.val**(other.val-1)) * self.jacobian + (self.val**other.val * np.log(np.abs(self.val)) * other.jacobian))
		except AttributeError:
			return AutoDiff(other**self.val, np.log(other) * other**self.val * self.der, self.n, 0, np.log(other) * other**self.val * self.jacobian)

	# Unary operations
	# Unary addition: identity
	def __pos__(self):
		return self

	# Unary subtration: negation
	def __neg__(self):
		# If  AutoDiff of same variable, values and derivatives should both just add
		return AutoDiff(self.val * -1, self.der * -1, self.n, 0, self.jacobian * -1)

	def __abs__(self):
		return AutoDiff(abs(self.val), ((self.val * self.der) / abs(self.val)),
				self.n, 0, ((self.val * self.jacobian) / abs(self.val)))

	def __invert__(self):
		return AutoDiff(~self.val, self.der * -1, self.n, 0, self.jacobian * -1)

	def __eq__(self, other):
		try:
			if np.all(np.equal(self.val,other.val)) and np.all(np.equal(self.der,other.der)):
				return True
			else:
				return False
		except AttributeError:
			return False

	def __ne__(self, other):
		try:
			if np.all(np.equal(self.val,other.val)) and np.all(np.equal(self.der,other.der)):
				return False
			else:
				return True
		except AttributeError:
			return True

def vectorize(ad_functions, n_inputs = 1, n_vectors = 1):
	'''
	INPUTS
	======
	ad_functions: 	a list or row vector of AutoDiff objects
	n_inputs:		number of input variables the final function will use
	n_vectors: 		length of vector for input variables

	RETURNS
	=======
	An AutoDiff object of vector functions with calculated value, derivative, and jacobian.

	EXAMPLES
	========
	>>> x = AutoDiff(3, np.array([[2, 0, 0]]), n=3, k=1)
	>>> y = AutoDiff(2, np.array([[0, 2, 0]]), n=3, k=2)
	>>> z = AutoDiff(-1, np.array([[0, 0, 2]]), n=3, k=3)
	>>> fs = [3*x + 2*y + 4*z, x - y + z, x/2, 2*x - 2*y]
	>>> f = vectorize(fs, 3)
	>>> f.val
	np.array([[9],[0],[1.5],[2]]
	>>> f.der
	np.array([[6, 4, 8], [2, -2, 2], [1, 0, 0], [4, -4, 0]])
	>>> f.jacobian
	np.array([[3, 2, 4], [1, -1, 1], [0.5, 0, 0], [2, -2, 0]])
	'''
	if n_vectors == 1:
		val = np.zeros([len(ad_functions), n_vectors])
		der = np.zeros([len(ad_functions), n_inputs])
		jacobian = np.zeros([len(ad_functions), n_inputs])
		for i, f in enumerate(ad_functions):
			val[i, :] = f.val
			der[i, :] = f.der
			jacobian[i, :] = f.jacobian
	else:
		val = np.zeros([len(ad_functions), n_vectors])
		der = np.zeros([n_vectors, len(ad_functions), n_inputs])
		jacobian = np.zeros([n_vectors, len(ad_functions), n_inputs])
		for j in range(n_vectors):
			der_j = der[j]
			jac_j = jacobian[j]
			for i, f in enumerate(ad_functions):
				val[i, :] = f.val.T
				der_j[i, :] = f.der[j]
				jac_j[i, :]	= f.jacobian[j]

	return AutoDiff(val, der, n_inputs, 0, jacobian)



import numpy as np

class Dual():
	def __init__(self, Real, Dual = 1):
		self.Real = Real
		self.Dual = Dual
		self.coefficients = []

	def makeHighestOrder(self, order):
		theLongdual = self._createNestedduals(self.Real, self.Dual, order)**1
		return theLongdual

	def _createNestedduals(self, value, dual, order):
		if order == 1:
			return Dual(value)
		else:
			return Dual(self._createNestedduals(value, dual, order - 1), dual)

	def buildCoefficients(self, n):
		coeffs = []
		for i in range(n + 1):
			theCoeff = self
			# get duals
			for j in range(i):
				theCoeff = theCoeff.Dual
			# get Reals
			for k in range(n-i):
				theCoeff = theCoeff.Real
			coeffs.append(theCoeff)
		self.coefficients = coeffs

	def __str__(self):
		if len(self.coefficients) == 0:
			return "{} + {}ε".format(self.Real, self.Dual)
		else:
			return "{}".format(self.coefficients)

	def __add__(self, other):
		try:
			return Dual(self.Real + other.Real, self.Dual + other.Dual)
		except AttributeError:
			return Dual(self.Real + other, self.Dual)

	def __radd__(self, other):
		try:
			return Dual(self.Real + other.Real, self.Dual + other.Dual)
		except AttributeError:
			return Dual(self.Real + other, self.Dual)

	def __sub__(self, other):
		try:
			return Dual(self.Real - other.Real, self.Dual - other.Dual)
		except AttributeError:
			return Dual(self.Real - other, self.Dual)

	def __rsub__(self, other):
		try:
			return Dual(self.Real - other.Real, self.Dual - other.Dual)
		except AttributeError:
			return Dual(other - self.Real, -1 * self.Dual)


	def __mul__(self, other):
		try:
			return Dual(self.Real * other.Real, self.Dual * other.Real + self.Real * other.Dual)
		except AttributeError:
			return Dual(self.Real * other, self.Dual * other)

	def __rmul__(self, other):
		try:
			return Dual(self.Real * other.Real, self.Dual * other.Real + self.Real * other.Dual)
		except AttributeError:
			return Dual(self.Real * other, self.Dual * other)

	def __truediv__(self, other):
		try:
			return Dual(self.Real / other.Real, (self.Dual*other.Real - self.Real*other.Dual) / other.Real**2)
		except AttributeError:
			return Dual(self.Real / other, (self.Dual / other))

	def __rtruediv__(self, other):
		try:
			return Dual(other.Real / self.Real, (other.Dual*self.Real - other.Real*self.Dual) / self.Real**2)
		except AttributeError:
			return Dual(other / self.Real, -1.0*((other*self.Dual)/(self.Real**2)))

	def __pow__(self, other):
		other = float(other) if type(other)==int else other
		try: # need to do
			return Dual(self.Real ** other.Real, self.Dual ** other.Dual)
		except AttributeError:
			return Dual(self.Real ** other, other * self.Dual * (self.Real ** (other - 1)))

	def __rpow__(self, other):
		try: # need to do
			return Dual(other.Real ** self.Real, other.Dual ** self.Dual)
		except AttributeError:
			return Dual(other ** self.Real, self.Dual * np.log(other) * (other ** self.Real))

	# Unary functions
	def __neg__(self):
		try:
			return Dual(-1.0 * self.Real, -1.0 * self.Dual)
		except:
			return -1.0 * self

	def __abs__(self):
		return Dual(abs(self.Real), self.Dual/abs(self.Real))

	def __pos__(self):
		try:
			return Dual(self.Real, self.Dual)
		except:
			return self

	# Comparison
	def __eq__(self, other):
		try:
			if np.all(self.Dual == other.Dual) and np.all(self.Real == other.Real):
				return True
			else:
				return False
		except AttributeError:
			return False

	def __ne__(self, other):
		try:
			if np.all(self.Dual == other.Dual) and np.all(self.Real == other.Real):
				return False
			else:
				return True
		except AttributeError:
			return True

def vectorizeDual(functions, order):
	n = len(functions)
	m = order + 1
	vector = np.zeros([n, m])
	for i, f in enumerate(functions):
   	 f.buildCoefficients(order)
   	 for j, coeff in enumerate(f.coefficients):
   		 vector[i,j] = coeff

	# Create a copy of one function
	f_everything = functions[0]
	f_everything.coefficients = np.zeros([n, m])
	for i in range(m):
   	 f_everything.coefficients[:, i] = vector[:, i]

	return f_everything


def makeHessianVars(x,y):
	xh = x.makeHighestOrder(3)
	yh = y.makeHighestOrder(3)
	return xh, yh


import warnings
import pytest
import numpy as np
import os,sys,inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0,parentdir)
from ADPYNE.AutoDiff import AutoDiff, vectorize
import ADPYNE.elemFunctions as ef

# helper function tests

def test_convertNonArray_array():
	AD = AutoDiff(np.array([[1,2]]),1)
	assert np.all(np.equal(AD.val, np.array([[1,2]])))

def test_convertNonArray_num():
	AD = AutoDiff(1,1)
	assert np.all(np.equal(AD.val, np.array([[1]])))

def test_calcJacobian_array():
	AD = AutoDiff(1,2)
	assert np.all(np.equal(AD.jacobian, np.array([[1]])))

def test_calcJacobian_array_withJ():
	AD = AutoDiff(1,1,1,0,np.array([[1]]))
	assert np.all(np.equal(AD.jacobian, np.array([[1]])))

def test_calcJacobian_vector():
	AD = AutoDiff(4, np.array([[2, 1]]).T, n=2, k=1)
	assert np.all(np.equal(AD.jacobian, np.array([[1, 0]])))
	AD = AutoDiff(3, np.array([[1, 2]]).T, n=2, k=2)
	assert np.all(np.equal(AD.jacobian, np.array([[0, 1]])))

def test_calcDerivative():
	AD = AutoDiff(4, 2, n=4, k=3)
	assert np.all(np.equal(AD.der, np.array([[0, 0, 2, 0]])))

# addition tests
def test_add_ad_results():
	# single input cases
	# positive numbers
	x = AutoDiff(5, 2)
	f = x + x
	assert f.val == 10
	assert f.der == 4
	assert f.jacobian == 2
	# negative numbers
	y = AutoDiff(-5, 2)
	f = y + y
	assert f.val == -10
	assert f.der == 4
	assert f.jacobian == 2

def test_add_vector_results():
	x = AutoDiff(np.array([[3],[1]]), np.array([[2, 1]]).T, 2, 1)
	y = AutoDiff(np.array([[2],[-3]]), np.array([[3, 2]]).T, 2, 2)
	f = x + y
	assert np.all(f.val == np.array([[5], [-2]]))
	assert np.all(f.der == np.array([[2, 3], [1, 2]]))
	assert np.all(f.jacobian == np.array([[1, 1], [1, 1]]))

def test_add_constant_results():
	# single input case
	# positive numbers
	x = AutoDiff(5, 2)
	f = x + 3
	assert f.val == 8
	assert f.der == 2
	assert f.jacobian == 1
	# negative numbers
	x = AutoDiff(-5, 2)
	f = x + 3
	assert f.val == -2
	assert f.der == 2
	assert f.jacobian == 1

def test_add_constant_vector_results():
	x = AutoDiff(np.array([[1, 3]]).T, np.array([[2, 1]]).T, 2, 1)
	f = x + 3
	assert np.all(f.val == np.array([[4, 6]]).T)
	assert np.all(f.der == np.array([[2, 0], [1, 0]]))
	assert np.all(f.jacobian == np.array([[1, 0], [1, 0]]))

# reverse addition tests
def test_radd_constant_results():
	# single input case
	# positive numbers
	x = AutoDiff(5, 2)
	f = 3 + x
	assert f.val == 8
	assert f.der == 2
	assert f.jacobian == 1
	# negative numbers
	x = AutoDiff(-5, 2)
	f = 3 + x
	assert f.val == -2
	assert f.der == 2
	assert f.jacobian == 1

def test_radd_constant_vector_results():
	x = AutoDiff(np.array([[1, 3]]).T, np.array([[2, 1]]).T, 2, 1)
	f = 3 + x
	assert np.all(f.val == np.array([[4, 6]]).T)
	assert np.all(f.der == np.array([[2, 0], [1, 0]]))
	assert np.all(f.jacobian == np.array([[1, 0], [1, 0]]))

# subtraction tests
def test_sub_ad_results():
	# single input cases
	# positive numbers
	x = AutoDiff(5, 2)
	f = x - x
	assert f.val == 0
	assert f.der == 0
	assert f.jacobian == 0
	# negative numbers
	y = AutoDiff(-5, 2)
	f = y - y
	assert f.val == 0
	assert f.der == 0
	assert f.jacobian == 0

def test_sub_vector_results():
	x = AutoDiff([3, 1], [2, 1], 2, 1)
	y = AutoDiff([2, -3], [3, 2], 2, 2)
	f = x - y
	assert np.all(f.val == np.array([[1], [4]]))
	assert np.all(f.der == np.array([[2, -3], [1, -2]]))
	assert np.all(f.jacobian == np.array([[1, -1], [1, -1]]))

def test_sub_constant_results():
	# single input case
	# positive numbers
	x = AutoDiff(5, 2)
	f = x - 3
	assert f.val == 2
	assert f.der == 2
	assert f.jacobian == 1
	# negative numbers
	x = AutoDiff(-5, 2)
	f = x - 3
	assert f.val == -8
	assert f.der == 2
	assert f.jacobian == 1

def test_sub_constant_vector_results():
	x = AutoDiff([1, 3], [2, 1], 2, 1)
	f = x - 3
	assert np.all(f.val == np.array([[-2, 0]]).T)
	assert np.all(f.der == np.array([[2, 0], [1, 0]]))
	assert np.all(f.jacobian == np.array([[1, 0], [1, 0]]))

# reverse subtraction tests
def test_rsub_constant_results():
	# single input case
	# positive numbers
	x = AutoDiff(5, 2)
	f = 3 - x
	assert f.val == -2
	assert f.der == 2
	assert f.jacobian == 1
	# negative numbers
	x = AutoDiff(-5, 2)
	f = 3 - x
	assert f.val == 8
	assert f.der == 2
	assert f.jacobian == 1

def test_rsub_constant_vector_results():
	x = AutoDiff([1, 3], [2, 1], 2, 1)
	f = 3 - x
	assert np.all(f.val == np.array([[2, 0]]).T)
	assert np.all(f.der == np.array([[2, 0], [1, 0]]))
	assert np.all(f.jacobian == np.array([[1, 0], [1, 0]]))

# multiplication tests
def test_mul_ad_results():
	# single input case
	# positive numbers
	x = AutoDiff(5, 2)
	f = x * x
	assert f.val == 25
	assert f.der == 20
	assert f.jacobian == 10
	# negative numbers
	x = AutoDiff(-5, 2)
	f = x * x
	assert f.val == 25
	assert f.der == -20
	assert f.jacobian == -10

def test_mul_vector_results():
	x = AutoDiff([3, 1], [2, 1], 2, 1)
	y = AutoDiff([2, -3], [1, 2], 2, 2)
	f = x*y
	assert np.all(f.val == np.array([[6, -3]]).T)
	assert np.all(f.der == np.array([[4, 3], [-3, 2]]))
	assert np.all(f.jacobian == np.array([[2, 3], [-3, 1]]))

def test_mul_constant_results():
	# single input case
	# positive numbers
	x = AutoDiff(5, 2)
	f = x * 3
	assert f.val == 15
	assert f.der == 6
	assert f.jacobian == 3
	# negative numbers
	x = AutoDiff(-5, 2)
	f = x * 3
	assert f.val == -15
	assert f.der == 6
	assert f.jacobian == 3

def test_mul_constant_vector_results():
	x = AutoDiff([3, 1], [2, 1], 2, 1)
	f = x * 3
	assert np.all(f.val == np.array([[9, 3]]).T)
	assert np.all(f.der == np.array([[6, 0], [3, 0]]))
	assert np.all(f.jacobian == np.array([[3, 0], [3, 0]]))

# reverse multiplication tests
def test_rmul_constant_results():
	# single input case
	# positive numbers
	x = AutoDiff(5, 2)
	f = 3 * x
	assert f.val == 15
	assert f.der == 6
	assert f.jacobian == 3
	# negative numbers
	x = AutoDiff(-5, 2)
	f = 3 * x
	assert f.val == -15
	assert f.der == 6
	assert f.jacobian == 3

def test_rmul_constant_vector_results():
	x = AutoDiff([3, 1], [2, 1], 2, 1)
	f = 3 * x
	assert np.all(f.val == np.array([[9, 3]]).T)
	assert np.all(f.der == np.array([[6, 0], [3, 0]]))
	assert np.all(f.jacobian == np.array([[3, 0], [3, 0]]))

# division tests
def test_truediv_ad_results():
	# single input case
	# positive numbers
	x = AutoDiff(5, 2)
	f = x / x
	assert f.val == 1
	assert f.der == 0
	assert f.jacobian == 0
	# negative numbers
	x = AutoDiff(-5, 2)
	f = x / x
	assert f.val == 1
	assert f.der == 0
	assert f.jacobian == 0

def test_truediv_vector_results():
	x = AutoDiff([3, 1], [2, 1], 2, 1)
	y = AutoDiff([2, -3], [1, 2], 2, 2)
	f = x/y
	assert np.all(f.val == np.array([[3/2, -1/3]]).T)
	assert np.all(f.der == np.array([[1, -0.75], [-1/3, -2/9]]))
	assert np.all(f.jacobian == np.array([[0.5, -0.75], [-1/3, -1/9]]))

def test_truediv_constant_results():
	# single input case
	# positive numbers
	x = AutoDiff(9, 6)
	f = x / 3
	assert f.val == 3
	assert f.der == 2
	assert f.jacobian == (1/3)
	# negative numbers
	x = AutoDiff(-9, 6)
	f = x / 3
	assert f.val == -3
	assert f.der == 2
	assert f.jacobian == (1/3)

def test_truediv_constant_vector_results():
	x = AutoDiff([-9, 3], [2, 1], 2, 1)
	f = x / 3
	assert np.all(f.val == np.array([[-3, 1]]).T)
	assert np.all(f.der == np.array([[2/3, 0], [1/3, 0]]))
	assert np.all(f.jacobian == np.array([[1/3, 0], [1/3, 0]]))

# reverse division tests
def test_rtruediv_constant_results():
	# single input case
	# positive numbers
	x = AutoDiff(3, 2)
	f = 6 / x
	assert f.val == 2
	assert f.der == -4/3
	assert f.jacobian == -2/3
	# negative numbers
	x = AutoDiff(-3, 2)
	f = 6 / x
	assert f.val == -2
	assert f.der == -4/3
	assert f.jacobian == -2/3

def test_rtruediv_constant_vector_results():
	x = AutoDiff([-9, 3], [2, 1], 1, 1)
	f = 3 / x
	assert np.all(f.val == np.array([[-1/3, 1]]).T)
	assert np.all(f.der == np.array([[-3*((-9)**(-2))*2], [-3*((3)**(-2))*1]]))
	assert np.all(f.jacobian == np.array([[-3*((-9)**(-2))*1], [-3*((3)**(-2))*1]]))

# power tests
def test_pow_ad_results():
	x = AutoDiff(2, 1)
	f = x**x
	assert f.val == 4
	assert f.der == 4 + np.log(16)
	assert f.jacobian == 4 + np.log(16)

def test_rpow_vector_results():
	x = AutoDiff([4, 3], [2, 1], 2, 1)
	y = AutoDiff([2, 1], [1, 3], 2, 2)
	f = x**y
	assert np.all(f.val == np.array([[4**2, 3**1]]).T)
	assert np.all(f.der == np.array([[2*(4**(2-1))*2, (4**2) * np.log(4) * 1], [1*(3**(1-1))*1, (3**1) * np.log(3)*3]]))
	assert np.all(f.jacobian == np.array([[2*(4**(2-1))*1, (4**2) * np.log(4) * 1], [1*(3**(1-1))*1, (3**1) * np.log(3)*1]]))

def test_pow_constant_results():
	# positive numbers
	x = AutoDiff(5, 2)
	f = x**3
	assert f.val == 125
	assert f.der == 150
	assert f.jacobian == 75
	# negative numbers
	x = AutoDiff(-5, 2)
	f = x**3
	assert f.val == -125
	assert f.der == 150
	assert f.jacobian == 75

def test_pow_constant_vector_results():
	x = AutoDiff([4, 3], [2, 1], 1, 1)
	f = x**3
	assert np.all(f.val == np.array([[4**3, 3**3]]).T)
	assert np.all(f.der == np.array([[3*(4**2)*2], [3*(3**2)*1]]))
	assert np.all(f.jacobian == np.array([[3*(4**2)*1], [3*(3**2)*1]]))

# reverse power tests
def test_rpow_constant_results():
	x = AutoDiff(5, 2)
	f = 3**x
	assert f.val == 243
	assert f.der == 486 * np.log(3)
	assert f.jacobian == 243 * np.log(3)

def test_rpow_constant_vector_results():
	x = AutoDiff([4, 3], [2, 1], 1, 1)
	f = 3**x
	assert np.all(f.val == np.array([[3**(4), 3**3]]).T)
	assert np.all(f.der == np.array([[(3**(4))*2 * np.log(3)], [(3**(3))*1 * np.log(3)]]))
	assert np.all(f.jacobian == np.array([[(3**(4))*1 * np.log(3)], [(3**(3))*1 * np.log(3)]]))


# positive tests
def test_pos_results():
	# positive numbers
	x = AutoDiff(5, 2)
	f = + x
	assert f.val == 5
	assert f.der == 2
	assert f.jacobian == 1
	# negative numbers
	y = AutoDiff(-5, 2)
	f = + y
	assert f.val == -5
	assert f.der == 2
	assert f.jacobian == 1

def test_pos_vector_results():
	x = AutoDiff([4, 3], [2, 1], 1, 1)
	f = + x
	assert np.all(f.val == np.array([[4, 3]]).T)
	assert np.all(f.der == np.array([[2], [1]]))
	assert np.all(f.jacobian == np.array([[1], [1]]))
	y = AutoDiff([-4, -3], [2, 1], 1, 1)
	g = + y
	assert np.all(g.val == np.array([[-4, -3]]).T)
	assert np.all(g.der == np.array([[2], [1]]))
	assert np.all(g.jacobian == np.array([[1], [1]]))

# negation tests
def test_neg_results():
	# positive numbers
	x = AutoDiff(5, 2)
	f = - x
	assert f.val == -5
	assert f.der == -2
	assert f.jacobian == -1
	# negative numbers
	y = AutoDiff(-5, 2)
	f = - y
	assert f.val == 5
	assert f.der == -2
	assert f.jacobian == -1

def test_neg_vector_results():
	x = AutoDiff([4, 3], [2, 1], 1, 1)
	f = - x
	assert np.all(f.val == np.array([[-4, -3]]).T)
	assert np.all(f.der == np.array([[-2], [-1]]))
	assert np.all(f.jacobian == np.array([[-1], [-1]]))
	y = AutoDiff([-4, -3], [2, 1], 1, 1)
	g = - y
	assert np.all(g.val == np.array([[4, 3]]).T)
	assert np.all(g.der == np.array([[-2], [-1]]))
	assert np.all(g.jacobian == np.array([[-1], [-1]]))

def test_neg_constant_results():
	x = 3
	f = - x
	assert f == -3

# absolute value tests
def test_abs_results():
	# positive numbers
	x = AutoDiff(5, 2)
	f = abs(x)
	assert f.val == 5
	assert f.der == 2
	assert f.jacobian == 1
	# negative numbers
	y = AutoDiff(-5, 2)
	f = abs(y)
	assert f.val == 5
	assert f.der == -2
	assert f.jacobian == -1

def test_abs_vector_results():
	x = AutoDiff([4, 3], [2, 1], 1, 1)
	f = abs(x)
	assert np.all(f.val == np.array([[4, 3]]).T)
	assert np.all(f.der == np.array([[2], [1]]))
	assert np.all(f.jacobian == np.array([[1], [1]]))
	y = AutoDiff([-4, -3], [2, 1], 1, 1)
	g = abs(y)
	assert np.all(g.val == np.array([[4, 3]]).T)
	assert np.all(g.der == np.array([[-2], [-1]]))
	assert np.all(g.jacobian == np.array([[-1], [-1]]))

def test_abs_constant_results():
	x = -3
	f = abs(x)
	assert f == 3

# Comparison tests
def test_eq_results():
	x = AutoDiff(5, 2)
	y = AutoDiff(5, 2)
	z = AutoDiff(5, 1)
	assert x == y
	assert (x == z) == False

def test_eq_vector_results():
	w = AutoDiff([4, 5], [2, 1], 1, 1)
	x = AutoDiff([4, 3], [2, 1], 1, 1)
	y = AutoDiff([4, 3], [2, 1], 1, 1)
	z = AutoDiff([4, 5], [2, 1], 1, 1)
	assert np.all(x == y)
	assert np.all(x==z) == False
	assert np.all(w==x) == False

def test_eq_constant():
	x = AutoDiff(5, 2)
	assert (x == 5) == False

def test_ne_results():
	x = AutoDiff(5, 2)
	y = AutoDiff(5, 2)
	z = AutoDiff(5, 1)
	assert x != z
	assert (x != y) == False

def test_neq_vector_results():
	w = AutoDiff([4, 5], [2, 1], 1, 1)
	x = AutoDiff([4, 3], [2, 1], 1, 1)
	y = AutoDiff([4, 3], [2, 1], 1, 1)
	z = AutoDiff([4, 5], [2, 1], 1, 1)
	assert np.all(x != y) == False
	assert np.all(x!=z)
	assert np.all(w!=x)

def test_ne_constant():
	x = AutoDiff(5, 2)
	assert x != 5

def test_vectorize():
	x = AutoDiff(3, np.array([[2]]), n=3, k=1)
	y = AutoDiff(2, np.array([[2]]), n=3, k=2)
	z = AutoDiff(-1, np.array([[2]]), n=3, k=3)
	fs = [3*x + 2*y + 4*z, x - y + z, x/2, 2*x - 2*y]
	f = vectorize(fs, 3)
	assert np.all(f.val == np.array([[9],[0],[1.5],[2]]))
	assert np.all(f.der == np.array([[6, 4, 8], [2, -2, 2], [1, 0, 0], [4, -4, 0]]))
	assert np.all(f.jacobian == np.array([[3, 2, 4], [1, -1, 1], [0.5, 0, 0], [2, -2, 0]]))



import warnings
import pytest
import numpy as np
import os,sys,inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0,parentdir)
from ADPYNE.Dual import Dual
import ADPYNE.elemFunctions as ef
from ADPYNE.Dual import Dual

# ------------SINE----------------#

def test_sin_ad_results():
	# Positive real numbers
	x = Dual(0.5, 2.0)
	f = ef.sin(x)
	assert f.Real == np.array([[np.sin(0.5)]])
	assert f.Dual == np.array([[np.cos(0.5)*2.0]])
	# Negative real numbers
	y = Dual(-0.5, 2.0)
	f = ef.sin(y)
	assert f.Real == np.array([[np.sin(-0.5)]])
	assert f.Dual == np.array([[np.cos(-0.5)*2.0]])
	# Zero
	z = Dual(0.0, 2.0)
	f = ef.sin(z)
	assert f.Real == np.array([[np.sin(0)]])
	assert f.Dual == np.array([[2.0]])

def test_sin_constant_results():
	a = ef.sin(5)
	assert a == np.sin(5)
	b = ef.sin(-5)
	assert b == np.sin(-5)
	c = ef.sin(0)
	assert c == np.sin(0)

def test_sin_types():
	with pytest.raises(TypeError):
		ef.sin('x')
	with pytest.raises(TypeError):
		ef.sin("1234")


# ------------COSINE----------------#

def test_cos_ad_results():
	# Positive real numbers
	x = Dual(0.5, 2.0)
	f = ef.cos(x)
	assert f.Real == np.array([[np.cos(0.5)]])
	assert f.Dual == np.array([[np.sin(0.5)*-2.0]])
	# Negative real numbers
	y = Dual(-0.5, 2.0)
	f = ef.cos(y)
	assert f.Real == np.array([[np.cos(-0.5)]])
	assert f.Dual == np.array([[np.sin(-0.5)*-2.0]])
	# Zero
	z = Dual(0.0, 2.0)
	f = ef.cos(z)
	assert f.Real == np.array([[np.cos(0)]])
	assert f.Dual == np.array([[0.0]])

def test_cos_constant_results():
	a = ef.cos(5)
	assert a == np.cos(5)
	b = ef.cos(-5)
	assert b == np.cos(-5)
	c = ef.cos(0)
	assert c == np.cos(0)

def test_cos_types():
	with pytest.raises(TypeError):
		ef.cos('x')
	with pytest.raises(TypeError):
		ef.cos("1234")


# ------------TANGENT----------------#

def test_tan_ad_results():
	# Defined Realue and Dualivative when cos(Real)!=0
	# Positive reals
	x = Dual(0.5, 2.0)
	f = ef.tan(x)
	assert f.Real == np.array([[np.tan(0.5)]])
	assert f.Dual == np.array([[2.0 / (np.cos(0.5)**2)]])
	# Negative reals
	y = Dual(-0.5, 2.0)
	f = ef.tan(y)
	assert f.Real == np.array([[np.tan(-0.5)]])
	assert f.Dual == np.array([[2.0 / (np.cos(-0.5)**2)]])
	# Zero
	z = Dual(0.0, 2.0)
	f = ef.tan(z)
	assert f.Real == np.array([[np.tan(0)]])
	assert f.Dual == np.array([[2.0]])

	# Undefined Value and Derivative when cos(Real)==0
	with pytest.warns(RuntimeWarning):
		h = Dual(np.pi/2, 1.0)
		f = ef.tan(h)
		print("HERE",f.Real)
		assert np.isnan(f.Real)
		assert np.isnan(f.Dual)

def test_tan_constant_results():
	a = ef.tan(5)
	assert a == np.tan(5)
	b = ef.tan(-5)
	assert b == np.tan(-5)
	c = ef.tan(0)
	assert c == np.tan(0)
	# Realue undefined when cos(Real)==0
	with pytest.warns(RuntimeWarning):
		d = ef.tan(np.pi/2)
		assert np.isnan(d)

def test_tan_types():
	with pytest.raises(TypeError):
		ef.tan('x')
	with pytest.raises(TypeError):
		ef.tan("1234")


# ------------ARC SINE----------------#

def test_arcsin_ad_results():
	# positive real numbers
	x = Dual(0.5, 2)
	f = ef.arcsin(x)
	assert f.Real == np.array([[np.arcsin(0.5)]])
	assert f.Dual == np.array([[2/np.sqrt(1-0.5**2)]])

	# out of bounds - undefined sqrt
	with pytest.warns(RuntimeWarning):
		y = Dual(-2, 2)
		f = ef.arcsin(y)
		assert np.isnan(f.Real)
		assert np.isnan(f.Dual)

	# out of bounds - div by zero
	with pytest.warns(RuntimeWarning):
		y = Dual(1, 2)
		f = ef.arcsin(y)
		assert f.Real == np.array([[np.arcsin(1)]])
		assert np.isinf(f.Dual)

	# zero
	z = Dual(0, 2)
	f = ef.arcsin(z)
	assert f.Real == np.array([[0.0]])
	assert f.Dual == np.array([[2.0]])

def test_arcsin_constant_results():
	a = ef.arcsin(0.7)
	assert a == np.arcsin(0.7)

	b = ef.arcsin(0)
	assert b == np.arcsin(0)

	with pytest.warns(RuntimeWarning):
		c = ef.arcsin(-5)
		assert np.isnan(c)

def test_arcsin_types():
	with pytest.raises(TypeError):
		ef.arcsin('x')
	with pytest.raises(TypeError):
		ef.arcsin("1234")


# ------------ARC COSINE----------------#

def test_arccos_ad_results():
	# positive real numbers
	x = Dual(0.5, 2)
	f = ef.arccos(x)
	assert f.Real == np.array([[np.arccos(0.5)]])
	assert f.Dual == np.array([[-2/np.sqrt(1-0.5**2)]])

	# out of bounds - negative sqrt
	with pytest.warns(RuntimeWarning):
		y = Dual(-2, 2)
		f = ef.arccos(y)
		assert np.isnan(f.Real)
		assert np.isnan(f.Dual)

	# out of bounds - divide by 0
	with pytest.warns(RuntimeWarning):
		y = Dual(1, 2)
		f = ef.arccos(y)
		assert f.Real == np.array([[np.arccos(1)]])
		assert np.isneginf(f.Dual)

	# zero
	z = Dual(0, 2)
	f = ef.arccos(z)
	assert f.Real == np.array([[np.arccos(0)]])
	assert f.Dual == np.array([[-2/np.sqrt(1-0**2)]])

def test_arccos_constant_results():
	a = ef.arccos(0.7)
	assert a == np.arccos(0.7)

	b = ef.arccos(0)
	assert b == np.arccos(0)

	with pytest.warns(RuntimeWarning):
		c = ef.arccos(-5)
		assert np.isnan(c)


def test_arccos_types():
	with pytest.raises(TypeError):
		ef.arccos('x')
	with pytest.raises(TypeError):
		ef.arccos("1234")


# ------------ARC TANGENT----------------#

def test_arctan_ad_results():
	# positive real numbers
	x = Dual(3, 2)
	f = ef.arctan(x)
	assert f.Real == np.array([[np.arctan(3)]])
	assert f.Dual == np.array([[2/(1+(3)**2)]])

	# negative real numbers
	y = Dual(-2, 2)
	f = ef.arctan(y)
	assert f.Real == np.array([[np.arctan(-2)]])
	assert f.Dual == np.array([[2/(1+(-2)**2)]])

	# zero
	z = Dual(0, 2)
	f = ef.arctan(z)
	assert f.Real == np.array([[np.arctan(0)]])
	assert f.Dual == np.array([[2/(1+(0)**2)]])


def test_arctan_constant_results():
	a = ef.arctan(0.7)
	assert a == np.arctan(0.7)
	b = ef.arctan(-5)
	assert b == np.arctan(-5)
	c = ef.arctan(0)
	assert c == np.arctan(0)

def test_arctan_types():
	with pytest.raises(TypeError):
		ef.arctan('x')
	with pytest.raises(TypeError):
		ef.arctan("1234")



# ------------HYPERBOLIC SINE----------------#
def test_sinh_results():
	X = Dual(0.5, 2)
	f = ef.sinh(X)
	assert f.Real == np.sinh(0.5)
	assert f.Dual == np.array([[2*np.cosh(0.5)]])

def test_sinh_constant_results():
	a = ef.sinh(0.5)
	assert a == np.sinh(0.5)

def test_sinh_types():
	with pytest.raises(TypeError):
		ef.sinh('x')
	with pytest.raises(TypeError):
		ef.sinh("1234")

# ------------HYPERBOLIC COS----------------#
def test_cosh_results():
	X = Dual(0.5, 2)
	f = ef.cosh(X)
	assert f.Real == np.cosh(0.5)
	assert f.Dual == np.array([[np.sinh(0.5)*2]])

def test_cosh_constant_results():
	a = ef.cosh(0.5)
	assert a == np.cosh(0.5)

def test_cosh_types():
	with pytest.raises(TypeError):
		ef.cosh('x')
	with pytest.raises(TypeError):
		ef.cosh("1234")

# ------------HYPERBOLIC TAN----------------#
def test_tanh_results():
	X = Dual(0.5, 2)
	f = ef.tanh(X)
	assert f.Real == np.tanh(0.5)
	assert f.Dual == np.array([[2/(np.cosh(0.5)**2)]])

def test_tanh_constant_results():
	a = ef.tanh(0.5)
	assert a == np.tanh(0.5)

def test_tanh_types():
	with pytest.raises(TypeError):
		ef.tanh('x')
	with pytest.raises(TypeError):
		ef.tanh("1234")


# ------------HYPERBOLIC ARC SINE----------------#

def test_arcsinh_ad_results():
	# positive real numbers
	x = Dual(1, 2)
	f = ef.arcsinh(x)
	assert f.Real == np.arcsinh(1)
	assert f.Dual == np.array([[((2)/np.sqrt((1)**2 + 1))]])
	
	# negative real numbers
	y = Dual(-1, 2)
	f = ef.arcsinh(y)
	assert f.Real == np.arcsinh(-1)
	assert f.Dual == np.array([[((2)/np.sqrt((-1)**2 + 1))]])
	
	# zero
	z = Dual(0, 2)
	f = ef.arcsinh(z)
	assert f.Real == np.arcsinh(0)
	assert f.Dual == np.array([[((2)/np.sqrt((0)**2 + 1))]])

def test_arcsinh_constant_results():
	a = ef.arcsinh(5)
	assert a == np.arcsinh(5)
	b = ef.arcsinh(-5)
	assert b == np.arcsinh(-5)
	c = ef.arcsinh(0)
	assert c == np.arcsinh(0)

def test_arcsinh_types():
	with pytest.raises(TypeError):
		ef.arcsinh('x')
	with pytest.raises(TypeError):
		ef.arcsinh("1234")

# ------------HYPERBOLIC ARC COSINE----------------#

def test_arccosh_ad_results():
	# Realue defined at positive real numbers x >= 1
	# Dualivative defined at positive real numbers x > 1
	x = Dual(1.1, 2)
	f = ef.arccosh(x)
	assert f.Real == np.arccosh(1.1)
	assert f.Dual == np.array([[((2)/np.sqrt((1.1)**2 - 1))]])
	
	# Realue defined at x = 1, Dualivative not defined
	with pytest.warns(RuntimeWarning):
		y = Dual(1, 2)
		f = ef.arccosh(y)
		assert np.isinf(f.Dual)
	
	# neither Realue nor Dualivative defined at x < 1
	with pytest.warns(RuntimeWarning):
		z = Dual(0, 2)
		f = ef.arccosh(z)
		assert np.isnan(f.Real)
		assert np.isnan(f.Dual)

def test_arccosh_constant_results():
	a = ef.arccosh(5)
	assert a == np.arccosh(5)
	# Realue not defined at x < 1
	with pytest.warns(RuntimeWarning):
		a = ef.arccosh(0.9)
		assert np.isnan(a)

def test_arccosh_types():
	with pytest.raises(TypeError):
		ef.arccosh('x')
	with pytest.raises(TypeError):
		ef.arccosh("1234")

# ------------HYPERBOLIC ARC TANGENT----------------#

def test_arctanh_ad_results():
	# Realue defined at real numbers (-1, 1)
	x = Dual(0.5, 2)
	f = ef.arctanh(x)
	assert f.Real == np.arctanh(0.5)
	assert f.Dual == np.array([[((2)/(1-(0.5)**2))]])
	
	y = Dual(-0.99999, 2)
	f = ef.arctanh(y)
	assert f.Real == np.arctanh(-0.99999)
	assert f.Dual == np.array([[((2)/(1-(-0.99999)**2))]])

	# test for real numbers not in (-1, 1)
	with pytest.warns(RuntimeWarning):
		z = Dual(-1, 2)
		f = ef.arctanh(z)
		assert np.isinf(f.Real)
		assert np.isinf(f.Dual)

	with pytest.warns(RuntimeWarning):
		z = Dual(10, 2)
		f = ef.arctanh(z)
		assert np.isnan(f.Real)
		assert f.Dual == np.array([[((2)/(1-(10)**2))]])

def test_arctanh_constant_results():
	a = ef.arctanh(0.99999)
	assert a == np.arctanh(0.99999)
	b = ef.arctanh(-0.001)
	assert b == np.arctanh(-0.001)
	with pytest.warns(RuntimeWarning):
		a = ef.arctanh(-1)
		assert np.isinf(a)
	with pytest.warns(RuntimeWarning):
		b = ef.arctanh(10)
		assert np.isnan(b)

def test_arctanh_types():
	with pytest.raises(TypeError):
		ef.arctanh('x')
	with pytest.raises(TypeError):
		ef.arctanh("1234")

# ---------------EXPONENTIAL----------------#

def test_exp_ad_results():
	# Realue defined at all real numbers
	# positive numbers
	x = Dual(10, 2)
	f = ef.exp(x)
	assert f.Real == np.exp(10)
	assert f.Dual == 2*np.exp(10)
	
	y = Dual(-5, 2)
	f = ef.exp(y)
	assert f.Real == np.exp(-5)
	assert f.Dual == 2*np.exp(-5)
	
	z = Dual(0, 2)
	f = ef.exp(z)
	assert f.Real == np.exp(0)
	assert f.Dual == 2*np.exp(0)

def test_exp_constant_results():
	a = ef.exp(0)
	assert a == np.exp(0)
	b = ef.exp(5)
	assert b == np.exp(5)
	c = ef.exp(-10)
	assert c == np.exp(-10)

def test_exp_types():
	with pytest.raises(TypeError):
		ef.exp('x')
	with pytest.raises(TypeError):
		ef.exp("1234")

# ---------------LOG----------------#
def test_log_results():
	# Realue defined at positive real numbers x > 0
	# Dualivative defined at real numbers x ≠ 0
	X = Dual(0.5, 2)
	f = ef.log(X)
	assert f.Real == np.log(0.5)
	assert f.Dual == np.array([[2/0.5]])

	# Dualivative not defined at x = 0
	Y = Dual(0, 2)
	f = ef.log(Y)
	assert np.isneginf(f.Real)
	assert np.isinf(f.Dual)

	# Realue not defined at x < 0, Dualivative defined
	with pytest.warns(RuntimeWarning):
		Y = Dual(-0.5, 2)
		f = ef.log(Y)
		assert np.isnan(f.Real)
		assert f.Dual == np.array([[2/-0.5]])

def test_log_constant_results():
	a = ef.log(0.5)
	assert a == np.log(0.5)
	with pytest.warns(RuntimeWarning):
		b = ef.log(0)
		assert np.isneginf(b)
	with pytest.warns(RuntimeWarning):
		b = ef.log(-0.5)
		assert np.isnan(b)

def test_log_types():
	with pytest.raises(TypeError):
		ef.log('x')
	with pytest.raises(TypeError):
		ef.log("1234")


# ---------------LOG10----------------#
def test_log10_results():
	# Realue defined at positive real numbers x > 0
	# Dualivative defined at positive real numbers x > 0
	X = Dual(0.5, 2)
	f = ef.log10(X)
	assert f.Real == np.log10(0.5)
	assert f.Dual == np.array([[(1/((0.5)*np.log(10)))*2]])

	# neither Realue nor Dualivative defined at x = 0
	with pytest.warns(RuntimeWarning):
		Y = Dual(0, 2)
		f = ef.log10(Y)
		assert np.isinf(f.Real)
		assert np.isinf(f.Dual)

	# Realue not defined at x < 0, Dualivative defined
	with pytest.warns(RuntimeWarning):
		Y = Dual(-0.5, 2)
		f = ef.log10(Y)
		assert np.isnan(f.Real)
		assert f.Dual == np.array([[(2/(-0.5*np.log(10)))]])

def test_log10_constant_results():
	a = ef.log10(0.5)
	print(a,"a")
	print(np.log10(0.5),"b")
	assert a == np.log10(0.5)
	with pytest.warns(RuntimeWarning):
		b = ef.log10(0)
		assert np.isinf(b)
	with pytest.warns(RuntimeWarning):
		b = ef.log10(-0.5)
		assert np.isnan(b)

def test_log10_types():
	with pytest.raises(TypeError):
		ef.log10('x')
	with pytest.raises(TypeError):
		ef.log10("1234")

# ---------------LOGBASE----------------#
def test_logbase_results():
	# Realue defined
	# 0<base<1 and x>1
	# base>1 and x>0
	# base=x, base≠0≠1
	# Dualivative defined at positive real numbers x > 0
	X = Dual(0.5, 2)
	f = ef.logbase(X,2)
	assert f.Real == np.log(0.5)/np.log(2)
	assert f.Dual == np.array([[(1/(0.5*np.log(2)))*2]])

	# Realue/Dual not defined at 0<base<1 and x<1
	with pytest.warns(RuntimeWarning):
		Y = Dual(-0.5, 2)
		f = ef.logbase(Y,0.9)
		assert np.isnan(f.Real)
		assert f.Dual == np.array([[(1/(-0.5*np.log(0.9)))*2]])

	# Realue/Dual not defined at base<1 and x>0
	with pytest.warns(RuntimeWarning):
		Y = Dual(1, 2)
		f = ef.logbase(Y,-10)
		assert np.isnan(f.Real)
		assert np.isnan(f.Dual)

	# Realue not defined at base>1 and x<=0
	with pytest.warns(RuntimeWarning):
		Y = Dual(-1, 2)
		f = ef.logbase(Y,10)
		assert np.isnan(f.Real)
		assert f.Dual == np.array([[(1/(-1*np.log(10)))*2]])

	# Realue not defined at x = 0 , where base = 0
	with pytest.warns(RuntimeWarning):
		Y = Dual(0, 2)
		f = ef.logbase(Y,0)
		assert np.isnan(f.Real)
		assert np.isnan(f.Dual)

	# Realue not defined at x = base , where base = 1
	with pytest.warns(RuntimeWarning):
		Y = Dual(1, 2)
		f = ef.logbase(Y,1)
		assert np.isnan(f.Real)
		assert np.isinf(f.Dual)

	# Realue not defined at x = base , where base = 0
	with pytest.warns(RuntimeWarning):
		Y = Dual(0, 2)
		f = ef.logbase(Y,0)
		assert np.isnan(f.Real)
		assert np.isnan(f.Dual)

def test_logbase_constant_results():
	a = ef.logbase(0.5,2)
	assert a == np.log(0.5)/np.log(2)
	with pytest.warns(RuntimeWarning):
		b = ef.logbase(0,2)
		assert np.isneginf(b)
	with pytest.warns(RuntimeWarning):
		b = ef.logbase(-0.5,2)
		assert np.isnan(b)

def test_logbase_types():
	with pytest.raises(TypeError):
		ef.logbase('x',3)
	with pytest.raises(TypeError):
		ef.logbase("1234",3)
	with pytest.raises(TypeError):
		ef.logbase(3,"x")
	with pytest.raises(TypeError):
		ef.logbase(3,"1234")

# ---------------SQUARE ROOT----------------#

def test_sqrt_ad_results():
	# Positive reals
	x = Dual(0.5, 2.0)
	f = ef.sqrt(x)
	assert f.Real == np.array([[np.sqrt(0.5)]])
	assert f.Dual == np.array([[0.5 * 0.5 ** (-0.5) * 2.0]])

	# Realue defined but Dualivative undefined when x == 0
	with pytest.warns(RuntimeWarning):
		y = Dual(0, 2)
		f = ef.sqrt(y)
		assert f.Real == np.array([[0]])
		assert np.isinf(f.Dual)

	# Realue and Dualivative undefined when x < 0
	with pytest.warns(RuntimeWarning):
		z = Dual(-0.5, 2)
		f = ef.sqrt(z)
		assert np.isnan(f.Real)
		assert np.isnan(f.Dual)

def test_sqrt_constant_results():
	a = ef.sqrt(5)
	assert a == np.sqrt(5)
	b = ef.sqrt(0)
	assert b == np.sqrt(0)
	# Realue undefined when x < 0
	with pytest.warns(RuntimeWarning):
		c = ef.sqrt(-5)
		assert np.isnan(c)

def test_sqrt_types():
	with pytest.raises(TypeError):
		ef.sqrt('x')
	with pytest.raises(TypeError):
		ef.sqrt("1234")


# ---------------LOGISTIC FUNC----------------#

def test_logistic_ad_results():
	# Positive reals
	x = Dual(0.5, 2.0)
	f = ef.logistic(x)
	assert f.Real == np.array([[1/(1+np.exp(-0.5))]])
	assert f.Dual == np.array([[2*np.exp(-0.5)/((1+np.exp(-0.5))**2)]])
	
def test_logistic_constant_results():
	a = ef.logistic(5)
	assert a == 1/(1+np.exp(-5))
	b = ef.logistic(0)
	assert b == 1/(1+np.exp(-0))

def test_logistic_types():
	with pytest.raises(TypeError):
		ef.logistic('x')
	with pytest.raises(TypeError):
		ef.logistic("1234")


import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="ADPYNE-207",
	version="0.1.4",
	author="Group 12 - CS 207",
	author_email="ptoroisaza@g.harvard.edu, nvanderklaauw@g.harvard.edu, emmali@college.harvard.edu, yaoweili@g.harvard.edu",
	description="Automatic Differentiation Package for Python",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/PYNE-AD/cs207-FinalProject",
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	python_requires='>=3.6',
	install_requires=[
		'numpy>=1.16.4',
		'pytest>=5.2.1'
	],
	tests_require=[
		'pytest>=5.2.1'
	]
)


import warnings
import pytest
import numpy as np
import os,sys,inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0,parentdir)
from ADPYNE.Dual import Dual, makeHessianVars
import ADPYNE.elemFunctions as ef
import ADPYNE.elemFunctions as ef
from ADPYNE.Hessian import Hessian

# higher order derivative and __str__ test
def test_higherorderprint():
	x = Dual(2, 1)
	assert str(x) == "2 + 1ε"
	y = x.makeHighestOrder(3)
	f = y
	f.buildCoefficients(3)
	assert f.coefficients == [2.0, 1.0, 0.0, -0.0]
	assert str(f) == "[2.0, 1.0, 0.0, -0.0]"

# addition tests
def test_add_dual_results():
	# single input cases
	# positive numbers
	x = Dual(5, 1)
	f = x + x
	assert f.Real == 10
	assert f.Dual == 2
	# negative numbers
	y = Dual(-5, 1)
	f = y + y
	assert f.Real == -10
	assert f.Dual == 2

def test_add_vector_results():
	x = Dual(np.array([[3],[1]]), np.array([[2, 1]]).T)
	y = Dual(np.array([[2],[-3]]), np.array([[3, 2]]).T)
	f = x + y
	assert np.all(f.Real == np.array([[5], [-2]]))
	assert np.all(f.Dual == np.array([[5], [3]]))

def test_add_constant_results():
	# single input case
	# positive numbers
	x = Dual(5, 1)
	f = x + 3
	assert f.Real == 8
	assert f.Dual == 1
	# negative numbers
	x = Dual(-5, 1)
	f = x + 3
	assert f.Real == -2
	assert f.Dual == 1

def test_add_constant_vector_results():
	x = Dual(np.array([[1, 3]]).T, np.array([[2, 1]]).T)
	f = x + 3
	assert np.all(f.Real == np.array([[4, 6]]).T)
	assert np.all(f.Dual == np.array([[2], [1]]))

# reverse addition tests
def test_radd_constant_results():
	# single input case
	# positive numbers
	x = Dual(5, 1)
	f = 3 + x
	assert f.Real == 8
	assert f.Dual == 1
	# negative numbers
	x = Dual(-5, 1)
	f = 3 + x
	assert f.Real == -2
	assert f.Dual == 1

def test_radd_constant_vector_results():
	x = Dual(np.array([[1, 3]]).T, np.array([[2, 1]]).T)
	f = 3 + x
	assert np.all(f.Real == np.array([[4, 6]]).T)
	assert np.all(f.Dual == np.array([[2], [1]]))

# subtraction tests
def test_sub_ad_results():
	# single input cases
	# positive numbers
	x = Dual(5, 2)
	f = x - x
	assert f.Real == 0
	assert f.Dual == 0
	# negative numbers
	y = Dual(-5, 2)
	f = y - y
	assert f.Real == 0
	assert f.Dual == 0

def test_sub_vector_results():
	x = Dual(np.array([[3, 1]]), np.array([[2, 1]]))
	y = Dual(np.array([[2, -3]]), np.array([[3, 2]]))
	f = x - y
	assert np.all(f.Real == np.array([[1, 4]]))
	assert np.all(f.Dual == np.array([[-1], [-1]]))

def test_sub_constant_results():
	# single input case
	# positive numbers
	x = Dual(5, 2)
	f = x - 3
	assert f.Real == 2
	assert f.Dual == 2
	# negative numbers
	x = Dual(-5, 2)
	f = x - 3
	assert f.Real == -8
	assert f.Dual == 2

def test_sub_constant_vector_results():
	x = Dual(np.array([[1, 3]]), np.array([[2, 1]]))
	f = x - 3
	assert np.all(f.Real == np.array([[-2, 0]]))
	assert np.all(f.Dual == np.array([[2, 1]]))

# reverse subtraction tests
def test_rsub_constant_results():
	# single input case
	# positive numbers
	x = Dual(5, 2)
	f = 3 - x
	assert f.Real == -2
	assert f.Dual == -2
	# negative numbers
	x = Dual(-5, 2)
	f = 3 - x
	assert f.Real == 8
	assert f.Dual == -2

def test_rsub_constant_vector_results():
	x = Dual(np.array([[1, 3]]), np.array([[2, 1]]))
	f = 3 - x
	assert np.all(f.Real == np.array([[2, 0]]))
	assert np.all(f.Dual == np.array([[-2, -1]]))

# multiplication tests
def test_mul_dual_results():
	# single input case
	# positive numbers
	x = Dual(5, 2)
	f = x * x
	assert f.Real == 25
	assert f.Dual == 20
	# negative numbers
	x = Dual(-5, 2)
	f = x * x
	assert f.Real == 25
	assert f.Dual == -20

def test_mul_vector_results():
	x = Dual(np.array([[3, 1]]), np.array([[2, 1]]))
	y = Dual(np.array([[2, -1]]), np.array([[1, 2]]))
	f = x*y
	assert np.all(f.Real == np.array([[6, -1]]))
	assert np.all(f.Dual == np.array([[7, 1]]))

def test_mul_constant_results():
	# single input case
	# positive numbers
	x = Dual(5, 2)
	f = x * 3
	assert f.Real == 15
	assert f.Dual == 6
	# negative numbers
	x = Dual(-5, 2)
	f = x * 3
	assert f.Real == -15
	assert f.Dual == 6

def test_mul_constant_vector_results():
	x = Dual(np.array([[3, 1]]), np.array([[2, 1]]))
	f = x * 3
	assert np.all(f.Real == np.array([[9, 3]]))
	assert np.all(f.Dual == np.array([[6, 3]]))

# reverse multiplication tests
def test_rmul_constant_results():
	# single input case
	# positive numbers
	x = Dual(5, 2)
	f = 3 * x
	assert f.Real == 15
	assert f.Dual == 6
	# negative numbers
	x = Dual(-5, 2)
	f = 3 * x
	assert f.Real == -15
	assert f.Dual == 6

def test_rmul_constant_vector_results():
	x = Dual(np.array([[3, 1]]), np.array([[2, 1]]))
	f = 3 * x
	assert np.all(f.Real == np.array([[9, 3]]))
	assert np.all(f.Dual == np.array([[6, 3]]))

# division tests
def test_truediv_dual_results():
	# single input case
	# positive numbers
	x = Dual(5, 2)
	f = x / x
	assert f.Real == 1
	assert f.Dual == 0
	# negative numbers
	x = Dual(-5, 2)
	f = x / x
	assert f.Real == 1
	assert f.Dual == 0

def test_truediv_vector_results():
	x = Dual(np.array([[3, 1]]), np.array([[2, 1]]))
	y = Dual(np.array([[2, -3]]), np.array([[1, 2]]))
	f = x/y
	assert np.all(f.Real == np.array([[3/2, -1/3]]))
	assert np.all(f.Dual == np.array([[1/4, -5/9]]))

def test_truediv_constant_results():
	# single input case
	# positive numbers
	x = Dual(9, 6)
	f = x / 3
	assert f.Real == 3
	assert f.Dual == 2
	# negative numbers
	x = Dual(-9, 6)
	f = x / 3
	assert f.Real == -3
	assert f.Dual == 2

def test_truediv_constant_vector_results():
	x = Dual(np.array([[9, 3]]), np.array([[2, 1]]))
	f = x / 3
	assert np.all(f.Real == np.array([[3, 1]]))
	assert np.all(f.Dual == np.array([[2/3, 1/3]]))

# reverse division tests
def test_rtruediv_constant_results():
	# single input case
	# positive numbers
	x = Dual(3, 2)
	f = 6 / x
	assert f.Real == 2
	assert f.Dual == -4/3
	# negative numbers
	x = Dual(-3, 2)
	f = 6 / x
	assert f.Real == -2
	assert f.Dual == -4/3

def test_rtruediv_constant_vector_results():
	x = Dual(np.array([[-9, 1]]), np.array([[2, 1]]))
	f = 3 / x
	assert np.all(f.Real == np.array([[-1/3, 3]]))
	assert np.all(f.Dual == np.array([[-2/27, -3]]))

# power tests
def test_pow_dual_results():
	x = Dual(2, 1)
	f = x**x
	assert f.Real == 4
	assert f.Dual == 1

def test_rpow_vector_results():
	x = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	y = Dual(np.array([[2, 1]]), np.array([[1, 3]]))
	f = x**y
	assert np.all(f.Real == np.array([[16, 3]]))
	assert np.all(f.Dual == np.array([[2, 1]]))

def test_pow_constant_results():
	# positive numbers
	x = Dual(5, 2)
	f = x**3
	assert f.Real == 125
	assert f.Dual == 150
	# negative numbers
	x = Dual(-5, 2)
	f = x**3
	assert f.Real == -125
	assert f.Dual == 150

def test_pow_constant_vector_results():
	x = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	f = x**3
	assert np.all(f.Real == np.array([[64, 27]]))
	assert np.all(f.Dual == np.array([[96, 27]]))

# reverse power tests
def test_rpow_constant_results():
	x = Dual(5, 2)
	f = 3**x
	assert f.Real == 243
	assert f.Dual == 486 * np.log(3)
test_rpow_constant_results()

def test_rpow_constant_vector_results():
	x = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	f = 3**x
	assert np.all(f.Real == np.array([[81, 27]]))
	assert np.all(f.Dual == np.array([[162*np.log(3), 27*np.log(3)]]))
test_rpow_constant_vector_results()

# positive tests
def test_pos_results():
	# positive numbers
	x = Dual(5, 2)
	f = + x
	assert f.Real == 5
	assert f.Dual == 2
	# negative numbers
	y = Dual(-5, 2)
	f = + y
	assert f.Real == -5
	assert f.Dual == 2

def test_pos_vector_results():
	x = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	f = + x
	assert np.all(f.Real == np.array([[4, 3]]))
	assert np.all(f.Dual == np.array([[2, 1]]))
	y = Dual(np.array([[-4, -3]]), np.array([[2, 1]]))
	g = + y
	assert np.all(g.Real == np.array([[-4, -3]]))
	assert np.all(g.Dual == np.array([[2, 1]]))

# negation tests
def test_neg_results():
	# positive numbers
	x = Dual(5, 2)
	f = - x
	assert f.Real == -5
	assert f.Dual == -2
	# negative numbers
	y = Dual(-5, 2)
	f = - y
	assert f.Real == 5
	assert f.Dual == -2

def test_neg_vector_results():
	x = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	f = - x
	assert np.all(f.Real == np.array([[-4, -3]]))
	assert np.all(f.Dual == np.array([[-2, -1]]))
	y = Dual(np.array([[-4, -3]]), np.array([[2, 1]]))
	g = - y
	assert np.all(g.Real == np.array([[4, 3]]))
	assert np.all(g.Dual == np.array([[-2, -1]]))

def test_neg_constant_results():
	x = 3
	f = - x
	assert f == -3

# absolute value tests
def test_abs_results():
	# positive numbers
	x = Dual(5, 2)
	f = abs(x)
	assert f.Real == 5
	assert f.Dual == 0.4
	# negative numbers
	y = Dual(-5, 2)
	f = abs(y)
	assert f.Real == 5
	assert f.Dual == 0.4

def test_abs_vector_results():
	x = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	f = abs(x)
	assert np.all(f.Real == np.array([[4, 3]]))
	assert np.all(f.Dual == np.array([[1/2, 1/3]]))
	y = Dual(np.array([[-4, -3]]), np.array([[2, 1]]))
	g = abs(y)
	assert np.all(g.Real == np.array([[4, 3]]))
	assert np.all(g.Dual == np.array([[1/2, 1/3]]))

def test_abs_constant_results():
	x = -3
	f = abs(x)
	assert f == 3
	y = 3
	f = abs(y)
	assert f == 3

# Comparison tests
def test_eq_results():
	x = Dual(5, 2)
	y = Dual(5, 2)
	z = Dual(5, 1)
	assert x == y
	assert (x == z) == False

def test_eq_vector_results():
	w = Dual(np.array([[4, 5]]), np.array([[2, 1]]))
	x = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	y = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	z = Dual(np.array([[4, 5]]), np.array([[2, 1]]))
	assert np.all(x != y) == False
	assert np.all(x==z) == False
	assert np.all(w==x) == False

def test_eq_constant():
	x = Dual(5, 2)
	assert (x == 5) == False

def test_ne_results():
	x = Dual(5, 2)
	y = Dual(5, 2)
	z = Dual(5, 1)
	assert x != z
	assert (x != y) == False

def test_neq_vector_results():
	w = Dual(np.array([[4, 5]]), np.array([[2, 1]]))
	x = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	y = Dual(np.array([[4, 3]]), np.array([[2, 1]]))
	z = Dual(np.array([[4, 5]]), np.array([[2, 1]]))
	assert np.all(x != y) == False
	assert np.all(x!=z)
	assert np.all(w!=x)

def test_ne_constant():
	x = Dual(5, 2)
	assert x != 5


def test_hessian():
	x = Dual(3, np.array([1,0]))
	y = Dual(1, np.array([0,1]))
	xh,yh = makeHessianVars(x,y)
	func = (xh**2)*(yh**3)
	f = Hessian(func)
	assert f.value == (3**2)*(1**3)
	assert np.all(f.firstDer == np.array([2*3*(1)**3, 3*(3**2)*(1)**2]))
	assert np.all(f.hessian == np.array([[2*(1)**2, 6*3*(1)**2],[6*3*(1)**2, 6*(3**2)*1]]))